
from cuteSV.cuteSV_Description import Generation_VCF_header
from math import log10
import numpy as np

err = 0.1
prior = float(1/3)
Genotype = ["0/0", "0/1", "1/1"]

def log10sumexp(log10_probs):
	# Normalization of Genotype likelihoods
	m = max(log10_probs)
	return m + log10(sum(pow(10.0, x-m) for x in log10_probs))

def normalize_log10_probs(log10_probs):
	# Adjust the Genotype likelihoods
	log10_probs = np.array(log10_probs)
	lse = log10sumexp(log10_probs)
	return np.minimum(log10_probs - lse, 0.0)

def rescale_read_counts(c0, c1, max_allowed_reads=100):
	"""Ensures that n_total <= max_allowed_reads, rescaling if necessary."""
	Total = c0 + c1
	if Total > max_allowed_reads:
		c0 = int(max_allowed_reads * float(c0/Total))
		c1 = max_allowed_reads - c0
	return c0, c1

def cal_GL(c0, c1):
	# Approximate adjustment of events with larger read depth
	c0, c1 = rescale_read_counts(c0, c1)
	# original genotype likelihood
	# ori_GL00 = np.float64(pow((1-err), c0)*pow(err, c1)*comb(c0+c1,c0)*(1-prior)/2)
	# ori_GL11 = np.float64(pow(err, c0)*pow((1-err), c1)*comb(c0+c1,c0)*(1-prior)/2)
	# ori_GL01 = np.float64(pow(0.5, c0+c1)*comb(c0+c1,c0)*prior)
	
	ori_GL00 = np.float64(pow((1-err), c0)*pow(err, c1)*(1-prior)/2)
	ori_GL11 = np.float64(pow(err, c0)*pow((1-err), c1)*(1-prior)/2)
	ori_GL01 = np.float64(pow(0.5, c0+c1)*prior)

	# normalized genotype likelihood
	prob = list(normalize_log10_probs([log10(ori_GL00), log10(ori_GL01), log10(ori_GL11)]))
	GL_P = [pow(10, i) for i in prob]
	PL = [int(np.around(-10*log10(i))) for i in GL_P]
	GQ = [int(-10*log10(GL_P[1] + GL_P[2])), int(-10*log10(GL_P[0] + GL_P[2])), int(-10*log10(GL_P[0] + GL_P[1]))]
	QUAL = abs(np.around(-10*log10(GL_P[0]), 1))

	return Genotype[prob.index(max(prob))], "%d,%d,%d"%(PL[0], PL[1], PL[2]), max(GQ), QUAL

def cal_CIPOS(std, num):
	pos = int(1.96 * std / num ** 0.5)
	return "-%d,%d"%(pos,pos)

def threshold_ref_count(num):
	if num <= 2:
		return 10*num
	elif 3 <= num <= 5:
		return 5*num 
	elif 6 <= num <= 15:
		return 4*num
	else:
		return 3*num

def count_coverage(chr, s, e, f, read_count, up_bound, itround):
	status = 0
	iteration = 0
	primary_num = 0
	for i in f.fetch(chr, s, e):
		iteration += 1
		if i.flag not in [0,16]:
			continue
		primary_num += 1
		if i.reference_start < s and i.reference_end > e:
			read_count.add(i.query_name)
			if len(read_count) >= up_bound:
				status = 1
				break
		if iteration >= itround:
			if float(primary_num/iteration) <= 0.2:
				status = 1
			else:
				status = -1
			break

	return status

def generate_output(args, semi_result, contigINFO, argv, ref_g):
	
	'''
	Generation of VCF format file.
	VCF version: 4.2
	'''

	# genotype_trigger = TriggerGT[args.genotype]

	svid = dict()
	svid["INS"] = 0
	svid["DEL"] = 0
	svid["BND"] = 0
	svid["DUP"] = 0
	svid["INV"] = 0

	file = open(args.output, 'w')
	Generation_VCF_header(file, contigINFO, args.sample, argv)
	file.write("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t%s\n"%(args.sample))

	for i in semi_result:
		if i[1] in ["DEL", "INS"]:
			if i[1] == "INS":
				cal_end = int(i[2]) + 1
			else:
				cal_end = int(i[2]) + 1 + abs(int(float(i[3])))
			info_list = "{PRECISION};SVTYPE={SVTYPE};SVLEN={SVLEN};END={END};CIPOS={CIPOS};CILEN={CILEN};RE={RE};RNAMES={RNAMES}".format(
				PRECISION = "IMPRECISE" if i[8] == "0/0" else "PRECISE", 
				SVTYPE = i[1], 
				SVLEN = i[3], 
				END = str(cal_end), 
				CIPOS = i[5], 
				CILEN = i[6], 
				RE = i[4],
				RNAMES = i[12] if args.report_readid else "NULL")
			if i[1] =="DEL":
				info_list += ";STRAND=+-"
			if i[11] == "." or i[11] == None:
				filter_lable = "PASS"
			else:
				filter_lable = "PASS" if float(i[11]) >= 5.0 else "q5"
			file.write("{CHR}\t{POS}\t{ID}\t{REF}\t{ALT}\t{QUAL}\t{PASS}\t{INFO}\t{FORMAT}\t{GT}:{DR}:{RE}:{PL}:{GQ}\n".format(
				CHR = i[0], 
				POS = str(int(i[2]) + 1), 
				ID = "cuteSV.%s.%d"%(i[1], svid[i[1]]),
				REF = str(ref_g[i[0]].seq[max(int(i[2])-1, 0)]) if i[1] == 'INS' else str(ref_g[i[0]].seq[max(int(i[2])-1, 0):int(i[2])-int(i[3])]),
				ALT = "%s"%(str(ref_g[i[0]].seq[max(int(i[2])-1, 0)])+i[13] if i[1] == 'INS' else str(ref_g[i[0]].seq[max(int(i[2])-1, 0)])), 
				INFO = info_list, 
				FORMAT = "GT:DR:DV:PL:GQ", 
				GT = i[8],
				DR = i[7],
				RE = i[4],
				PL = i[9],
				GQ = i[10],
				QUAL = i[11],
				PASS = filter_lable))
			svid[i[1]] += 1
		elif i[1] == "DUP":
			cal_end = int(i[2]) + 1 + abs(int(float(i[3])))
			info_list = "{PRECISION};SVTYPE={SVTYPE};SVLEN={SVLEN};END={END};RE={RE};STRAND=-+;RNAMES={RNAMES}".format(
				PRECISION = "IMPRECISE" if i[6] == "0/0" else "PRECISE", 
				SVTYPE = i[1], 
				SVLEN = i[3], 
				END = str(cal_end), 
				RE = i[4],
				RNAMES = i[10] if args.report_readid else "NULL")
			if i[9] == ".":
				filter_lable = "PASS"
			else:
				filter_lable = "PASS" if float(i[9]) >= 5.0 else "q5"
			file.write("{CHR}\t{POS}\t{ID}\t{REF}\t{ALT}\t{QUAL}\t{PASS}\t{INFO}\t{FORMAT}\t{GT}:{DR}:{RE}:{PL}:{GQ}\n".format(
				CHR = i[0], 
				POS = str(int(i[2]) + 1), 
				ID = "cuteSV.%s.%d"%(i[1], svid[i[1]]),
				REF = str(ref_g[i[0]].seq[int(i[2])]),
				ALT = "<%s>"%(i[1]), 
				INFO = info_list, 
				FORMAT = "GT:DR:DV:PL:GQ", 
				GT = i[6],
				DR = i[5],
				RE = i[4],
				PL = i[7],
				GQ = i[8],
				QUAL = i[9],
				PASS = filter_lable))
			svid[i[1]] += 1
		elif i[1] == "INV":
			cal_end = int(i[2]) + 1 + abs(int(float(i[3])))
			info_list = "{PRECISION};SVTYPE={SVTYPE};SVLEN={SVLEN};END={END};RE={RE};STRAND={STRAND};RNAMES={RNAMES}".format(
				PRECISION = "IMPRECISE" if i[6] == "0/0" else "PRECISE", 
				SVTYPE = i[1], 
				SVLEN = i[3], 
				END = str(cal_end), 
				RE = i[4],
				STRAND = i[7],
				RNAMES = i[11] if args.report_readid else "NULL")
			if i[10] == ".":
				filter_lable = "PASS"
			else:
				filter_lable = "PASS" if float(i[10]) >= 5.0 else "q5"
			file.write("{CHR}\t{POS}\t{ID}\t{REF}\t{ALT}\t{QUAL}\t{PASS}\t{INFO}\t{FORMAT}\t{GT}:{DR}:{RE}:{PL}:{GQ}\n".format(
				CHR = i[0], 
				POS = str(int(i[2]) + 1), 
				ID = "cuteSV.%s.%d"%(i[1], svid[i[1]]),
				REF = str(ref_g[i[0]].seq[int(i[2])]),
				ALT = "<%s>"%(i[1]), 
				INFO = info_list, 
				FORMAT = "GT:DR:DV:PL:GQ", 
				GT = i[6],
				DR = i[5],
				RE = i[4],
				PL = i[8],
				GQ = i[9],
				QUAL = i[10],
				PASS = filter_lable))
			svid[i[1]] += 1
		else:
			# BND
			# info_list = "{PRECISION};SVTYPE={SVTYPE};CHR2={CHR2};END={END};RE={RE};RNAMES={RNAMES}".format(
			info_list = "{PRECISION};SVTYPE={SVTYPE};RE={RE};RNAMES={RNAMES}".format(
				PRECISION = "IMPRECISE" if i[7] == "0/0" else "PRECISE", 
				SVTYPE = "BND", 
				# CHR2 = i[3], 
				# END = str(int(i[4]) + 1), 
				RE = i[5],
				RNAMES = i[11] if args.report_readid else "NULL")
			if i[10] == ".":
				filter_lable = "PASS"
			else:
				filter_lable = "PASS" if float(i[10]) >= 5.0 else "q5"
			file.write("{CHR}\t{POS}\t{ID}\t{REF}\t{ALT}\t{QUAL}\t{PASS}\t{INFO}\t{FORMAT}\t{GT}:{DR}:{RE}:{PL}:{GQ}\n".format(
				CHR = i[0], 
				POS = str(int(i[2]) + 1), 
				ID = "cuteSV.%s.%d"%("BND", svid["BND"]), 
				REF = 'N',
				ALT = i[1], 
				INFO = info_list, 
				FORMAT = "GT:DR:DV:PL:GQ", 
				GT = i[7],
				DR = i[6],
				RE = i[5],
				PL = i[8],
				GQ = i[9],
				QUAL = i[10],
				PASS = filter_lable))
			svid["BND"] += 1


def generate_pvcf(args, result, contigINFO, argv):
	file = open(args.output, 'w')
	Generation_VCF_header(file, contigINFO, args.sample, argv)
	file.write("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t%s\n"%(args.sample))
	for i in result:
		if i == []:
			continue
		if i[3] == 'INS':
			info_list = "{PRECISION};SVTYPE={SVTYPE};SVLEN={SVLEN};END={END};CIPOS={CIPOS};CILEN={CILEN};RE={RE};RNAMES={RNAMES};".format(
				PRECISION = "IMPRECISE" if i[2] == "0/0" else "PRECISE", 
				SVTYPE = i[3], 
				SVLEN = i[4], 
				END = i[1], 
				CIPOS = str(i[6][0]) + ',' + str(i[6][1]), 
				CILEN = str(i[7][0]) + ',' + str(i[7][1]), 
				RE = i[8][0],
				RNAMES = i[9] if args.report_readid else "NULL")
		elif i[3] == 'DEL':
			info_list = "{PRECISION};SVTYPE={SVTYPE};SVLEN={SVLEN};END={END};CIPOS={CIPOS};CILEN={CILEN};RE={RE};RNAMES={RNAMES};STRAND=+-".format(
				PRECISION = "IMPRECISE" if i[2] == "0/0" else "PRECISE", 
				SVTYPE = i[3], 
				SVLEN = i[4], 
				END = i[1] + abs(i[4]), 
				CIPOS = str(i[6][0]) + ',' + str(i[6][1]), 
				CILEN = str(i[7][0]) + ',' + str(i[7][1]), 
				RE = i[8][0],
				RNAMES = i[9] if args.report_readid else "NULL")
		elif i[3] == 'DUP':
			info_list = "{PRECISION};SVTYPE={SVTYPE};SVLEN={SVLEN};END={END};RE={RE};RNAMES={RNAMES};STRAND=-+".format(
				PRECISION = "IMPRECISE" if i[2] == "0/0" else "PRECISE", 
				SVTYPE = i[3], 
				SVLEN = i[4], 
				END = i[5], 
				RE = i[8][0],
				RNAMES = i[9] if args.report_readid else "NULL")
		elif i[3] == 'INV':
			info_list = "{PRECISION};SVTYPE={SVTYPE};SVLEN={SVLEN};END={END};RE={RE};RNAMES={RNAMES};STRAND={STRAND}".format(
				PRECISION = "IMPRECISE" if i[2] == "0/0" else "PRECISE", 
				SVTYPE = i[3], 
				SVLEN = i[4], 
				END = i[5], 
				RE = i[8][0],
				RNAMES = i[9] if args.report_readid else "NULL",
				STRAND = i[14])
		else:
			# BND
			info_list = "{PRECISION};SVTYPE={SVTYPE};RE={RE};RNAMES={RNAMES};".format(
					PRECISION = "IMPRECISE" if i[2] == "0/0" else "PRECISE", 
					SVTYPE = i[3], 
					RE = i[8][0],
					RNAMES = i[9] if args.report_readid else "NULL")
		if i[13] == "." or i[13] == None:
			filter_lable = "PASS"
		else:
			filter_lable = "PASS" if float(i[13]) >= 5.0 else "q5"
		file.write("{CHR}\t{POS}\t{ID}\t{REF}\t{ALT}\t{QUAL}\t{PASS}\t{INFO}\t{FORMAT}\t{GT}:{DR}:{RE}:{PL}:{GQ}\n".format(
			CHR = i[0], 
			POS = str(i[1]), 
			ID = i[10],
			REF = i[11],
			ALT = i[12], 
			QUAL = '.' if i[13] == None else i[13],
			PASS = filter_lable,
			INFO = info_list, 
			FORMAT = "GT:DR:DV:PL:GQ", 
			GT = i[2],
			DR = i[8][1],
			RE = i[8][0],
			PL = i[8][2],
			GQ = i[8][3]
			))

			
def load_valuable_chr(path):
	valuable_chr = dict()
	valuable_chr["DEL"] = list()
	valuable_chr["DUP"] = list()
	valuable_chr["INS"] = list()
	valuable_chr["INV"] = list()
	valuable_chr["TRA"] = dict()

	for svtype in ["DEL", "DUP", "INS", "INV"]:
		file = open("%s%s.sigs"%(path, svtype), 'r')
		for line in file:
			chr = line.strip('\n').split('\t')[1]
			if chr not in valuable_chr[svtype]:
				valuable_chr[svtype].append(chr)
		file.close()
		valuable_chr[svtype].sort()

	file = open("%s%s.sigs"%(path, "TRA"), 'r')
	for line in file:
		chr1 = line.strip('\n').split('\t')[1]
		chr2 = line.strip('\n').split('\t')[4]
		
		if chr1 not in valuable_chr["TRA"]:
			valuable_chr["TRA"][chr1] = list()
		if chr2 not in valuable_chr["TRA"][chr1]:
			valuable_chr["TRA"][chr1].append(chr2)

	file.close()
	for chr1 in valuable_chr["TRA"]:
		valuable_chr["TRA"][chr1].sort()

	return valuable_chr




import sys
import numpy as np
from collections import Counter
from cuteSV.cuteSV_genotype import cal_GL, cal_CIPOS, threshold_ref_count, count_coverage
import time

'''
*******************************************
				TO DO LIST
*******************************************
	1. Identify DP with samfile pointer;
	2. Add CIPOS, CILEN and/or CIEND;
	3. Determine (IM)PRECISE type.
*******************************************

'''

def resolution_DEL(path, chr, svtype, read_count, threshold_gloab, max_cluster_bias,
				 minimum_support_reads, bam_path, action, gt_round):

	'''
	cluster DEL
	********************************************************************************************
	path:	DEL.sigs
	chr:	chromosome id
	svtype:	<DEL>
	
	SEQTYPE		read_count 	max_cluster_bias 	sv_size		threshold_gloab 	threshold_local 
	--------------------------------------------------------------------------------------------
	CCS			3			200 bp (<500 bp)	30 bp 		0.4					0.5
	CLR			5/10		200 bp (<500 bp)	50 bp 		0.3					0.7
	--------------------------------------------------------------------------------------------
	
	Input file format
	--------------------------------------------------------------------------------------------
	column	#1	#2	#3	#4	#5
			DEL	CHR	BP	LEN	ID	
	#1	deletion type
	#2	chromosome number
	#3	breakpoint in each read
	#4	DEL_len in each read
	#5	read ID
	********************************************************************************************
	'''

	semi_del_cluster = list()
	semi_del_cluster.append([0,0,''])
	candidate_single_SV = list()

	file = open(path, 'r')
	for line in file:
		seq = line.strip('\n').split('\t')
		if seq[1] != chr:
			continue

		pos = int(seq[2])
		indel_len = int(seq[3])
		read_id = seq[4]
		
		if pos - semi_del_cluster[-1][0] > max_cluster_bias:
			if len(semi_del_cluster) >= read_count:
				if semi_del_cluster[-1][0] == semi_del_cluster[-1][1] == 0:
					pass
				else:
					generate_del_cluster(semi_del_cluster, 
										chr, 
										svtype, 
										read_count, 
										threshold_gloab, 
										# threshold_local, 
										minimum_support_reads, 
										candidate_single_SV,
										bam_path,
										max_cluster_bias,
										action,
										gt_round)
			semi_del_cluster = []
			semi_del_cluster.append([pos, indel_len, read_id])
		else:
			semi_del_cluster.append([pos, indel_len, read_id])

	if len(semi_del_cluster) >= read_count:
		if semi_del_cluster[-1][0] == semi_del_cluster[-1][1] == 0:
			pass
		else:
			generate_del_cluster(semi_del_cluster, 
								chr, 
								svtype, 
								read_count, 
								threshold_gloab, 
								# threshold_local, 
								minimum_support_reads, 
								candidate_single_SV,
								bam_path,
								max_cluster_bias,
								action,
								gt_round)
	file.close()
	return candidate_single_SV

def generate_del_cluster(semi_del_cluster, chr, svtype, read_count, 
	threshold_gloab, minimum_support_reads, candidate_single_SV, 
	bam_path, max_cluster_bias, action, gt_round):

	'''
	generate deletion
	*************************************************************
	threshold_gloab 	threshold_local 	minimum_support_reads
	-------------------------------------------------------------
		0.3					0.7 					5		CLR
		0.4					0.5 				  <=5		CCS
	*************************************************************
	'''

	# Remove duplicates
	read_tag = dict()
	for element in semi_del_cluster:
		if element[2] not in read_tag:
			read_tag[element[2]] = element
		else:
			if element[1] > read_tag[element[2]][1]:
				read_tag[element[2]] = element

	if len(read_tag) < read_count:
		return

	read_tag2SortedList = sorted(list(read_tag.values()), key = lambda x:x[1])
	global_len = [i[1] for i in read_tag2SortedList]
	DISCRETE_THRESHOLD_LEN_CLUSTER_DEL_TEMP = threshold_gloab * np.mean(global_len)

	last_len = read_tag2SortedList[0][1]

	allele_collect = list()
	'''
	*************************************************************
		#1 				#2			#3			#4
	-------------------------------------------------------------
		del-breakpoint	del-len		#support 	read-id
	*************************************************************
	'''
	allele_collect.append([[read_tag2SortedList[0][0]],[read_tag2SortedList[0][1]],[],
		[read_tag2SortedList[0][2]]])

	for i in read_tag2SortedList[1:]:
		if i[1] - last_len > DISCRETE_THRESHOLD_LEN_CLUSTER_DEL_TEMP:
			allele_collect[-1][2].append(len(allele_collect[-1][0]))
			allele_collect.append([[],[],[],[]])

		allele_collect[-1][0].append(i[0])
		allele_collect[-1][1].append(i[1])
		allele_collect[-1][3].append(i[2])
		last_len = i[1]
	allele_collect[-1][2].append(len(allele_collect[-1][0]))
	allele_sort = sorted(allele_collect, key = lambda x:x[2])

	for allele in allele_sort:
		if allele[2][0] >= minimum_support_reads:
			breakpointStart = np.mean(allele[0])
			search_threshold = np.min(allele[0])
			CIPOS = cal_CIPOS(np.std(allele[0]), len(allele[0]))
			signalLen = np.mean(allele[1])
			signalLen_STD = np.std(allele[1])
			CILEN = cal_CIPOS(np.std(allele[1]), len(allele[1]))

			if action:
				DV, DR, GT, GL, GQ, QUAL = call_gt(bam_path, 
													int(search_threshold), 
													chr, 
													allele[3], 
													max_cluster_bias, 
													gt_round)
			else:
				DR = '.'
				GT = './.'
				GL = '.,.,.'
				GQ = "."
				QUAL = "."
			candidate_single_SV.append([chr, 
										svtype, 
										str(int(breakpointStart)), 
										str(int(-signalLen)), 
										str(allele[2][0]), 
										str(CIPOS),
										str(CILEN),
										str(DR),
										str(GT),
										str(GL),
										str(GQ),
										str(QUAL),
										str(','.join(allele[3]))])
	

def resolution_INS(path, chr, svtype, read_count, threshold_gloab, 
	max_cluster_bias, minimum_support_reads, bam_path, action, gt_round):
	
	'''
	cluster INS
	********************************************************************************************
	path:	INS.sigs
	chr:	chromosome id
	svtype:	<INS>
	
	SEQTYPE		read_count 	max_cluster_bias 	sv_size		threshold_gloab 	threshold_local 
	--------------------------------------------------------------------------------------------
	CCS			3			200 bp (<500 bp)	30 bp 		0.65				0.7
	CLR			5/10		100 bp (<500 bp)	50 bp 		0.2					0.6
	--------------------------------------------------------------------------------------------
	
	Input file format
	--------------------------------------------------------------------------------------------
	column	#1	#2	#3	#4	#5
			INS	CHR	BP	LEN	ID	
	#1	insertion type
	#2	chromosome number
	#3	breakpoint in each read
	#4	DEL_len in each read
	#5	read ID
	********************************************************************************************
	'''

	semi_ins_cluster = list()
	semi_ins_cluster.append([0,0,'',''])
	candidate_single_SV = list()

	file = open(path, 'r')
	for line in file:
		seq = line.strip('\n').split('\t')
		if seq[1] != chr:
			continue

		pos = int(seq[2])
		indel_len = int(seq[3])
		read_id = seq[4]
		ins_seq = seq[5]
		
		if pos - semi_ins_cluster[-1][0] > max_cluster_bias:
			if len(semi_ins_cluster) >= read_count:
				if semi_ins_cluster[-1][0] == semi_ins_cluster[-1][1] == 0:
					pass
				else:
					generate_ins_cluster(semi_ins_cluster, 
										chr, 
										svtype, 
										read_count, 
										threshold_gloab, 
										# threshold_local, 
										minimum_support_reads, 
										candidate_single_SV,
										bam_path,
										max_cluster_bias,
										action,
										gt_round)
			semi_ins_cluster = []
			semi_ins_cluster.append([pos, indel_len, read_id, ins_seq])
		else:
			semi_ins_cluster.append([pos, indel_len, read_id, ins_seq])

	if len(semi_ins_cluster) >= read_count:
		if semi_ins_cluster[-1][0] == semi_ins_cluster[-1][1] == 0:
			pass
		else:
			generate_ins_cluster(semi_ins_cluster, 
								chr, 
								svtype, 
								read_count, 
								threshold_gloab, 
								# threshold_local, 
								minimum_support_reads, 
								candidate_single_SV,
								bam_path,
								max_cluster_bias,
								action,
								gt_round)
	file.close()
	return candidate_single_SV

def generate_ins_cluster(semi_ins_cluster, chr, svtype, read_count, 
	threshold_gloab, minimum_support_reads, candidate_single_SV, 
	bam_path, max_cluster_bias, action, gt_round):
		
	'''
	generate deletion
	*************************************************************
	threshold_gloab 	threshold_local 	minimum_support_reads
	-------------------------------------------------------------
		0.2					0.6 					5		CLR
		0.65				0.7 				  <=5		CCS
	*************************************************************
	'''

	# Remove duplicates
	read_tag = dict()
	for element in semi_ins_cluster:
		if element[2] not in read_tag:
			read_tag[element[2]] = element
		else:
			if element[1] > read_tag[element[2]][1]:
				read_tag[element[2]] = element

	if len(read_tag) < read_count:
		return

	read_tag2SortedList = sorted(list(read_tag.values()), key = lambda x:x[1])
	# start&end breakpoint
	global_len = [i[1] for i in read_tag2SortedList]
	DISCRETE_THRESHOLD_LEN_CLUSTER_INS_TEMP = threshold_gloab * np.mean(global_len)

	last_len = read_tag2SortedList[0][1]

	allele_collect = list()
	allele_collect.append([[read_tag2SortedList[0][0]],
							[read_tag2SortedList[0][1]],
							[], 
							[read_tag2SortedList[0][2]],
							[read_tag2SortedList[0][3]]])

	for i in read_tag2SortedList[1:]:
		if i[1] - last_len > DISCRETE_THRESHOLD_LEN_CLUSTER_INS_TEMP:
			allele_collect[-1][2].append(len(allele_collect[-1][0]))
			allele_collect.append([[],[],[],[],[]])

		allele_collect[-1][0].append(i[0])
		allele_collect[-1][1].append(i[1])
		allele_collect[-1][3].append(i[2])
		allele_collect[-1][4].append(i[3])
		last_len = i[1]
	allele_collect[-1][2].append(len(allele_collect[-1][0]))
	allele_sort = sorted(allele_collect, key = lambda x:x[2])

	for allele in allele_sort:
		if allele[2][0] >= minimum_support_reads:
			breakpointStart = np.mean(allele[0])
			CIPOS = cal_CIPOS(np.std(allele[0]), len(allele[0]))
			signalLen = np.mean(allele[1])
			signalLen_STD = np.std(allele[1])
			CILEN = cal_CIPOS(np.std(allele[1]), len(allele[1]))
			ideal_ins_seq = '<INS>'
			for i in allele[4]:
				if len(i) >= int(signalLen):
					ideal_ins_seq = i[0:int(signalLen)]
					break
			if ideal_ins_seq == '<INS>':
				continue
				
			if action:
				DV, DR, GT, GL, GQ, QUAL = call_gt(bam_path, 
													int(breakpointStart), 
													chr, 
													allele[3], 
													# max_cluster_bias, 
													1000,
													gt_round)
			else:
				DR = '.'
				GT = './.'
				GL = '.,.,.'
				GQ = "."
				QUAL = "."
			candidate_single_SV.append([chr, 
										svtype, 
										str(int(breakpointStart)), 
										str(int(signalLen)), 
										str(allele[2][0]), 
										str(CIPOS),
										str(CILEN),
										str(DR),
										str(GT),
										str(GL),
										str(GQ),
										str(QUAL),
										str(','.join(allele[3])),
										ideal_ins_seq])


def run_del(args):
	return resolution_DEL(*args)

def run_ins(args):
	return resolution_INS(*args)

def call_gt(bam_path, search_threshold, chr, read_id_list, max_cluster_bias, gt_round):
	import pysam
	querydata = set()
	bamfile = pysam.AlignmentFile(bam_path)
	search_start = max(int(search_threshold) - max_cluster_bias, 0)
	search_end = min(int(search_threshold) + max_cluster_bias, bamfile.get_reference_length(chr))

	up_bound = threshold_ref_count(len(read_id_list))

	status = count_coverage(chr, 
							search_start, 
							search_end, 
							bamfile, 
							querydata, 
							up_bound, 
							gt_round)
	bamfile.close()

	if status == -1:
		DR = '.'
		GT = "./."
		GL = ".,.,."
		GQ = "."
		QUAL = "."

	# elif status == 1:
	# 	pass
	else:
		DR = 0
		for query in querydata:
			if query not in read_id_list:
				DR += 1
		GT, GL, GQ, QUAL = cal_GL(DR, len(read_id_list))
	return len(read_id_list), DR, GT, GL, GQ, QUAL



import sys
import numpy as np
from collections import Counter
from cuteSV.cuteSV_genotype import cal_GL, threshold_ref_count, count_coverage

'''
*******************************************
				TO DO LIST
*******************************************
	1. Identify DP with samfile pointer;
	2. Add CIPOS, CILEN and/or CIEND;
	3. Determine (IM)PRECISE type.
	4. Filter DUP to improve INS FN rate.
*******************************************
'''

def resolution_DUP(path, chr, read_count, max_cluster_bias, sv_size, 
	bam_path, action, MaxSize, gt_round):
	semi_dup_cluster = list()
	semi_dup_cluster.append([0, 0, ''])
	candidate_single_SV = list()

	file = open(path, 'r')
	for line in file:
		seq = line.strip('\n').split('\t')
		if seq[1] != chr:
			continue

		pos_1 = int(seq[2])
		pos_2 = int(seq[3])
		read_id = seq[4]
		
		if pos_1 - semi_dup_cluster[-1][0] > max_cluster_bias or pos_2 - semi_dup_cluster[-1][1] > max_cluster_bias:
			if len(semi_dup_cluster) >= read_count:
				if semi_dup_cluster[-1][0] == semi_dup_cluster[-1][1] == 0:
					pass
				else:
					generate_dup_cluster(semi_dup_cluster, 
										chr, 
										read_count, 
										max_cluster_bias, 
										sv_size, 
										candidate_single_SV,
										bam_path,
										action,
										MaxSize,
										gt_round)
			semi_dup_cluster = []
			semi_dup_cluster.append([pos_1, pos_2, read_id])
		else:
			semi_dup_cluster.append([pos_1, pos_2, read_id])

	if len(semi_dup_cluster) >= read_count:
		if semi_dup_cluster[-1][0] == semi_dup_cluster[-1][1] == 0:
			pass
		else:
			generate_dup_cluster(semi_dup_cluster, 
								chr, 
								read_count, 
								max_cluster_bias, 
								sv_size, 
								candidate_single_SV,
								bam_path,
								action,
								MaxSize,
								gt_round)
	file.close()
	return candidate_single_SV

def generate_dup_cluster(semi_dup_cluster, chr, read_count, max_cluster_bias, 
	sv_size, candidate_single_SV, bam_path, action, MaxSize, gt_round):
	# calculate support reads
	support_read = list(set([i[2] for i in semi_dup_cluster]))
	if len(support_read) < read_count:
		return

	low_b = int(len(semi_dup_cluster)*0.4)
	up_b = int(len(semi_dup_cluster)*0.6)

	if low_b == up_b:
		breakpoint_1 = semi_dup_cluster[low_b][0]
		breakpoint_2 = semi_dup_cluster[low_b][1]
	else:
		breakpoint_1 = [i[0] for i in semi_dup_cluster[low_b:up_b]]
		breakpoint_2 = [i[1] for i in semi_dup_cluster[low_b:up_b]]
		breakpoint_1 = int(sum(breakpoint_1)/len(semi_dup_cluster[low_b:up_b]))
		breakpoint_2 = int(sum(breakpoint_2)/len(semi_dup_cluster[low_b:up_b]))


	if sv_size <= breakpoint_2 - breakpoint_1 <= MaxSize:
		if action:
			import time
			# time_start = time.time()
			DV, DR, GT, GL, GQ, QUAL = call_gt(bam_path, 
												breakpoint_1, 
												breakpoint_2, 
												chr, 
												support_read, 
												min(max_cluster_bias, breakpoint_2 - breakpoint_1),
												gt_round)
			# print(DV, DR, GT, GL, GQ, QUAL)
			# cost_time = time.time() - time_start
			# print("DUP", chr, int(breakpoint_1), int(breakpoint_2), DR, DV, QUAL, "%.4f"%cost_time)
		else:
			DR = '.'
			GT = './.'
			GL = '.,.,.'
			GQ = "."
			QUAL = "."
		candidate_single_SV.append([chr,
									'DUP', 
									str(breakpoint_1), 
									str(breakpoint_2 - breakpoint_1), 
									str(len(support_read)),
									str(DR),
									str(GT),
									str(GL),
									str(GQ),
									str(QUAL),
									str(','.join(support_read))])


def run_dup(args):
	return resolution_DUP(*args)

def call_gt(bam_path, pos_1, pos_2, chr, read_id_list, max_cluster_bias, gt_round):
	import pysam
	bamfile = pysam.AlignmentFile(bam_path)
	querydata = set()
	search_start = max(int(pos_1 - max_cluster_bias/2), 0)
	search_end = min(int(pos_1 + max_cluster_bias/2), bamfile.get_reference_length(chr))

	up_bound = threshold_ref_count(len(read_id_list))
	status = count_coverage(chr, 
							search_start, 
							search_end, 
							bamfile, 
							querydata, 
							up_bound, 
							gt_round)

	if status == -1:
		DR = '.'
		GT = "./."
		GL = ".,.,."
		GQ = "."
		QUAL = "."

	elif status == 1:
		DR = 0
		for query in querydata:
			if query not in read_id_list:
				DR += 1
		GT, GL, GQ, QUAL = cal_GL(DR, len(read_id_list))

	else:
		search_start = max(int(pos_2 - max_cluster_bias/2), 0)
		search_end = min(int(pos_2 + max_cluster_bias/2), bamfile.get_reference_length(chr))
		status_2 = count_coverage(chr, 
									search_start, 
									search_end, 
									bamfile, 
									querydata, 
									up_bound, 
									gt_round)
		# status_2 judgement
		DR = 0
		for query in querydata:
			if query not in read_id_list:
				DR += 1
		GT, GL, GQ, QUAL = cal_GL(DR, len(read_id_list))

	bamfile.close()
	return len(read_id_list), DR, GT, GL, GQ, QUAL



from cuteSV.cuteSV_resolveINV import call_gt as call_gt_inv
from cuteSV.cuteSV_resolveTRA import call_gt as call_gt_tra
from cuteSV.cuteSV_resolveINDEL import call_gt as call_gt_indel
from cuteSV.cuteSV_resolveDUP import call_gt as call_gt_dup
from multiprocessing import Pool, Manager
import vcf
from pysam import VariantFile
import math
import time
import numpy as np
import logging


class Para(object):
    def __init__(self, record):
        self.chrom = record.chrom
        self.pos = parse_to_int(record.pos)
        self.svlen = parse_to_int(record.info['SVLEN']) if 'SVLEN' in record.info else 0
        self.end = parse_to_int(record.stop)
        self.cipos = record.info['CIPOS'] if 'CIPOS' in record.info else ('.', '.')
        self.ciend = record.info['CIEND'] if 'CIEND' in record.info else ('.', '.')
        self.id = record.id
        self.ref = record.ref
        self.alts = record.alts[0]
        self.qual = record.qual


def parse_to_int(sth):
    if sth == None:
        return 0
    elif isinstance(sth, str):
        return int(sth)
    elif isinstance(sth, list):
        return parse_to_int(sth[0])
    elif isinstance(sth, tuple):
        return parse_to_int(sth[0])
    elif isinstance(sth, int):
        return sth
    else:
        return sth


def parse_record(record):
    sv_type = record.info['SVTYPE']
    if sv_type == 'BND':
        sv_type = 'TRA'
    chrom1 = record.chrom
    start = parse_to_int(record.pos)
    if (record.info['SVTYPE'] == 'INS' or record.info['SVTYPE'] == 'DEL') and 'SVLEN' in record.info:  ###
        end = abs(parse_to_int(record.info['SVLEN']))
    else:
        try:
            end = parse_to_int(record.stop)
        except:
            pass   
    if sv_type == 'TRA':
        tra_alt = str(record.alts[0])
        if tra_alt[0] == 'N':
            if tra_alt[1] == '[':
                tra_type = 'A'
            else:
                tra_type = 'B'
        elif tra_alt[0] == '[':
            tra_type = 'C'
        else:
            tra_type = 'D'
        if tra_alt[0] == 'N':
            tra_alt = tra_alt[2:-1]
        else:
            tra_alt = tra_alt[1:-2]
        chrom2 = tra_alt.split(':')[0]
        end = int(tra_alt.split(':')[1])
    strand = '.'
    if sv_type != 'TRA':
        chrom2 = record.chrom
        if 'STRAND' in record.info:
            strand = record.info['STRAND']
        elif 'STRANDS' in record.info:
            strand = record.info['STRANDS']
    return sv_type, chrom1, chrom2, start, end, strand


def parse_sigs(var_type, work_dir):
    var_dict = dict()  #var_dict[chrom] = [chrom, start, end, read_id]
    with open(work_dir + var_type + '.sigs', 'r') as f:
        for line in f:
            seq = line.strip().split('\t')
            if seq[1] in var_dict:
                var_dict[seq[1]].append([seq[1], int(seq[2]), int(seq[3]), seq[4]])
            else:
                var_dict[seq[1]] = []
                var_dict[seq[1]].append([seq[1], int(seq[2]), int(seq[3]), seq[4]])
    return var_dict


def parse_invsigs(work_dir):
    var_dict = dict() # var_dict[chrom][strand] = [chrom, start, end, read_id]
    with open(work_dir + 'INV.sigs', 'r') as f:
        for line in f:
            seq = line.strip().split('\t')
            chrom = seq[1]
            strand = seq[2]
            if chrom in var_dict:
                if strand in var_dict[chrom]:
                    var_dict[chrom][strand].append([chrom, int(seq[3]), int(seq[4]), seq[5]])
                else:
                    var_dict[chrom][strand] = [[chrom, int(seq[3]), int(seq[4]), seq[5]]]
            else:
                var_dict[chrom] = dict()
                var_dict[chrom][strand] = [[chrom, int(seq[3]), int(seq[4]), seq[5]]]
    return var_dict


def parse_trasigs(work_dir):
    var_dict = dict()  #var_dict[chrom1][chrom2] = [[chrom2, start, end, read_id]]
    with open(work_dir + 'TRA.sigs', 'r') as f:
        for line in f:
            seq = line.strip().split('\t')
            chrom1 = seq[1]
            tra_type = seq[2]
            pos1 = int(seq[3])
            chrom2 = seq[4]
            pos2 = int(seq[5])
            read_id = seq[6]
            '''
            if chrom1 in var_dict:
                if tra_type in var_dict[chrom1]:
                    if chrom2 in var_dict[chrom1][tra_type]:
                        var_dict[chrom1][tra_type][chrom2].append([chrom2, pos1, pos2, read_id])
                    else:
                        var_dict[chrom1][tra_type][chrom2] = [[chrom2, pos1, pos2, read_id]]
                else:
                    var_dict[chrom1][tra_type] = dict()
                    var_dict[chrom1][tra_type][chrom2] = [[chrom2, pos1, pos2, read_id]]

            else:
                var_dict[chrom1] = dict()
                var_dict[chrom1][tra_type] = dict()
                var_dict[chrom1][tra_type][chrom2] = [[chrom2, pos1, pos2, read_id]]
            '''
            if chrom1 in var_dict:
                if chrom2 in var_dict[chrom1]:
                    var_dict[chrom1][chrom2].append([chrom2, pos1, pos2, read_id])
                else:
                    var_dict[chrom1][chrom2] = [[chrom2, pos1, pos2, read_id]]

            else:
                var_dict[chrom1] = dict()
                var_dict[chrom1][chrom2] = [[chrom2, pos1, pos2, read_id]]
    for chr1 in var_dict:
        for chr2 in var_dict[chr1]:
            var_dict[chr1][chr2].sort(key=lambda x:x[1])
    return var_dict


def check_same_variant(sv_type, end1, end2):
    if sv_type == 'INS' or sv_type == 'DEL':
        return 0.7 < min(end1, end2) / max(end1, end2) <= 1
    return abs(end1 - end2) < 1000


# var_list read_id which is similar to (pos, sv_end), the reads support the variant
def find_in_list(var_type, var_list, bias, pos, sv_end):
    '''
    if interval_start == 44058795:
        print('start find in list')
        print('interval_start=%d, interval_end=%d, pos=%d, sv_end=%d'%(interval_start, interval_end, pos, sv_end))
    '''
    if len(var_list) == 0:
        return [], 0
    left = 0
    right = len(var_list) - 1
    mid = 0
    while left < right:
        mid = (left + right) >> 1
        if var_list[mid][1] < pos:
            left = mid + 1
        else:
            right = mid
    read_id_list = set()
    search_start = 0
    search_end = 0
    if right > 0 and pos - var_list[right - 1][1] <= 1000:
        for i in range(right - 1, -1, -1):
            if check_same_variant(var_type, var_list[i][2], sv_end):
                read_id_list.add(var_list[i][3])
                search_start = var_list[i][1]
            if i > 0 and var_list[i][1] - var_list[i - 1][1] > bias:
                break
    if var_list[right][1] - pos <= 1000:
        for i in range(right, len(var_list)):
            if check_same_variant(var_type, var_list[i][2], sv_end):  # if abs(var_list[i][2] - sv_end) < 1000:
                read_id_list.add(var_list[i][3])
                search_end = var_list[i][1]
            if i < len(var_list) - 1 and var_list[i + 1][1] - var_list[i][1] > bias:
                break
    search_threshold = min(abs(pos - search_start), abs(pos - search_end))
    return list(read_id_list), search_threshold


def find_in_indel_list(var_type, var_list, bias, pos, sv_end, threshold_gloab):
    if len(var_list) == 0:
        return [], 0
    left = 0
    right = len(var_list) - 1
    mid = 0
    while left < right:
        mid = (left + right) >> 1
        if var_list[mid][1] < pos:
            left = mid + 1
        else:
            right = mid
    candidates = []
    if right > 0 and pos - var_list[right - 1][1] <= 1000:
        for i in range(right - 1, -1, -1):
            candidates.append(var_list[i])
            if i > 0 and var_list[i][1] - var_list[i - 1][1] > bias and pos - var_list[i - 1][1] > 1000:
                break
    if var_list[right][1] - pos <= 1000:
        for i in range(right, len(var_list)):
            candidates.append(var_list[i])
            if i < len(var_list) - 1 and var_list[i + 1][1] - var_list[i][1] > bias and var_list[i + 1][1] - pos > 1000:
                break
    if len(candidates) == 0:
        return [], 0
    read_tag = dict()
    for element in candidates:
        if element[3] not in read_tag:
            read_tag[element[3]] = element
        else:
            if element[2] > read_tag[element[3]][2]:
                read_tag[element[3]] = element
    read_tag2SortedList = sorted(list(read_tag.values()), key = lambda x:x[2])
    global_len = [i[2] for i in read_tag2SortedList]
    DISCRETE_THRESHOLD_LEN_CLUSTER_TEMP = threshold_gloab * np.mean(global_len)

    last_len = read_tag2SortedList[0][2]
    allele_collect = list()
    allele_collect.append([[read_tag2SortedList[0][1]],  # start
                            [read_tag2SortedList[0][2]],  # len
                            [],  # support
                            [read_tag2SortedList[0][3]]])  # read_id
    for i in read_tag2SortedList[1:]:
        if i[2] - last_len > DISCRETE_THRESHOLD_LEN_CLUSTER_TEMP:
            allele_collect[-1][2].append(len(allele_collect[-1][0]))
            allele_collect.append([[],[],[],[]])

        allele_collect[-1][0].append(i[1])
        allele_collect[-1][1].append(i[2])
        allele_collect[-1][3].append(i[3])
        last_len = i[2]
    allele_collect[-1][2].append(len(allele_collect[-1][0]))
    allele_sort = sorted(allele_collect, key = lambda x:x[2])

    allele_idx = 0
    nearest_gap = 0x3f3f3f3f
    for i in range(len(allele_sort)):
        allele = allele_sort[i]
        signalLen = np.mean(allele[1])
        if abs(signalLen - sv_end) < nearest_gap:
            allele_idx = i
            nearest_gap = abs(signalLen - sv_end)
    read_id_set = set(allele_sort[allele_idx][3])
    search_start = min(allele_sort[allele_idx][0])
    search_end = max(allele_sort[allele_idx][0])
    search_threshold = min(abs(pos - search_start), abs(pos - search_end))
    return list(read_id_set), search_threshold
        

def call_gt_wrapper(call_gt_args, gt_list, idx, row_count, para, strands, var_type):
    rname_list = []
    if var_type == 'INS' or var_type == 'DEL':
        gt_re, DR, genotype, GL, GQ, QUAL = call_gt_indel(*call_gt_args)
        rname_list = call_gt_args[3]
    if var_type == 'DUP':
        gt_re, DR, genotype, GL, GQ, QUAL = call_gt_dup(*call_gt_args)
        rname_list = call_gt_args[4]
    if var_type == 'INV':
        gt_re, DR, genotype, GL, GQ, QUAL = call_gt_inv(*call_gt_args)
        rname_list = call_gt_args[4]
    if var_type == 'TRA':
        gt_re, DR, genotype, GL, GQ, QUAL = call_gt_tra(*call_gt_args)
        rname_list = call_gt_args[5]
    rname = ','.join(rname_list)
    if rname == '':
        rname = 'NULL'
    gt_list[idx] = [para.chrom,
                    para.pos,
                    genotype,
                    var_type,
                    para.svlen,
                    para.end,
                    para.cipos,
                    para.ciend,
                    [gt_re, DR, GL, GQ, QUAL],
                    rname,
                    para.id,
                    para.ref, 
                    para.alts,
                    para.qual,
                    strands
    ]
    if idx > 0 and idx % 5000 == 0:
        logging.info(str(math.floor(idx / row_count * 100)) + '% SV calls of the given vcf has been processed.')


def force_calling(bam_path, ivcf_path, output_path, sigs_dir, max_cluster_bias_dict, threshold_gloab_dict, gt_round, threads):
    logging.info('Check the parameter -Ivcf: OK.')
    logging.info('Enable to perform force calling.')
    #print(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
    sv_dict = dict()
    #'''
    for sv_type in ["DEL", "INS", "DUP"]:
        sv_dict[sv_type] = parse_sigs(sv_type, sigs_dir)
    sv_dict['INV'] = parse_invsigs(sigs_dir)
    sv_dict['TRA'] = parse_trasigs(sigs_dir)
    #'''
    vcf_reader = VariantFile(ivcf_path, 'r')
    row_count = 0
    for record in vcf_reader.fetch():
        row_count += 1
    idx = -1
    gt_list = Manager().list([[] for x in range(row_count)])
    result = []
    process_pool = Pool(processes = threads)
    vcf_reader = VariantFile(ivcf_path, 'r')
    for record in vcf_reader.fetch():
        idx += 1
        sv_type, chrom, sv_chr2, pos, sv_end, sv_strand = parse_record(record)
        search_id_list = []
        if sv_type == 'TRA' and 'TRA' in sv_dict and chrom in sv_dict['TRA'] and sv_chr2 in sv_dict['TRA'][chrom]:
            search_id_list = sv_dict['TRA'][chrom][sv_chr2]
        elif sv_type == 'INV' and 'INV' in sv_dict and chrom in sv_dict['INV']:
            if sv_strand in sv_dict['INV'][chrom]:
                search_id_list = sv_dict['INV'][chrom][sv_strand]
            else:
                for strand_iter in sv_dict['INV'][chrom]:
                    sv_strand = strand_iter
                    search_id_list = sv_dict['INV'][chrom][strand_iter]
                    break
        elif sv_type != 'TRA' and sv_type != 'INV' and sv_type in sv_dict and chrom in sv_dict[sv_type]:
            search_id_list = sv_dict[sv_type][chrom]
        max_cluster_bias = 0
        if sv_type == 'INS' or sv_type == 'DEL':
            read_id_list, max_cluster_bias = find_in_indel_list(sv_type, search_id_list, max_cluster_bias_dict[sv_type], pos, sv_end, threshold_gloab_dict[sv_type])
        else:
            read_id_list, max_cluster_bias = find_in_list(sv_type, search_id_list, max_cluster_bias_dict[sv_type], pos, sv_end)
        if sv_type == 'INV' and len(read_id_list) == 0:
            for strand_iter in sv_dict['INV'][chrom]:
                if strand_iter != sv_strand:
                    search_id_list = sv_dict['INV'][chrom][strand_iter]
                    read_id_list, max_cluster_bias = find_in_list(sv_type, search_id_list, max_cluster_bias_dict[sv_type], pos, sv_end)
                    if len(read_id_list) != 0:
                        sv_strand = strand_iter
                        break
        #print(read_id_list)
        if sv_type == 'INS':
            max_cluster_bias = max(1000, max_cluster_bias)
        else:
            max_cluster_bias = max(max_cluster_bias_dict[sv_type], max_cluster_bias)

        para = Para(record)
        '''
        if sv_type == 'INS':
            call_gt_wrapper([bam_path, pos, chrom, read_id_list, max_cluster_bias, gt_round], gt_list, idx, row_count, para, sv_strand, 'INS')
        if sv_type == 'DEL':
            call_gt_wrapper([bam_path, pos, chrom, read_id_list, max_cluster_bias, gt_round], gt_list, idx, row_count, para, sv_strand, 'DEL')
        if sv_type == 'INV':
            call_gt_wrapper([bam_path, pos, sv_end, chrom, read_id_list, max_cluster_bias, gt_round], gt_list, idx, row_count, para, sv_strand, 'INV')
        if sv_type == 'DUP':
            call_gt_wrapper([bam_path, pos, sv_end, chrom, read_id_list, max_cluster_bias, gt_round], gt_list, idx, row_count, para, sv_strand, 'DUP')
        if sv_type == 'TRA':
            call_gt_wrapper([bam_path, pos, sv_end, chrom, sv_chr2, read_id_list, max_cluster_bias, gt_round], gt_list, idx, row_count, para, sv_strand, 'TRA')
        '''
        #'''
        if sv_type == 'INS':
            process_pool.apply_async(call_gt_wrapper, 
                args=([bam_path, pos, chrom, read_id_list, max_cluster_bias, gt_round], gt_list, idx, row_count, para, sv_strand, 'INS'))
        if sv_type == 'DEL':
            process_pool.apply_async(call_gt_wrapper, 
                args=([bam_path, pos, chrom, read_id_list, max_cluster_bias, gt_round], gt_list, idx, row_count, para, sv_strand, 'DEL'))
        if sv_type == 'INV':
            process_pool.apply_async(call_gt_wrapper, 
                args=([bam_path, pos, sv_end, chrom, read_id_list, max_cluster_bias, gt_round], gt_list, idx, row_count, para, sv_strand, 'INV'))
        if sv_type == 'DUP':
            process_pool.apply_async(call_gt_wrapper, 
                args=([bam_path, pos, sv_end, chrom, read_id_list, max_cluster_bias, gt_round], gt_list, idx, row_count, para, sv_strand, 'DUP'))
        if sv_type == 'TRA':
            process_pool.apply_async(call_gt_wrapper, 
                args=([bam_path, pos, sv_end, chrom, sv_chr2, read_id_list, max_cluster_bias, gt_round], gt_list, idx, row_count, para, sv_strand, 'TRA'))
        #'''
    process_pool.close()
    process_pool.join()
    logging.info('Finished force calling.')
    return gt_list


def run_fc(args):
    return force_calling(*args)




import sys
import numpy as np
from cuteSV.cuteSV_genotype import cal_GL, threshold_ref_count, count_coverage

def resolution_INV(path, chr, svtype, read_count, max_cluster_bias, sv_size, 
	bam_path, action, MaxSize, gt_round):
	'''
	cluster INV
	************************************************************************
	path:	INV.sigs
	chr:	chromosome id
	svtype:	<INV>
	
	SEQTYPE		read_count 	max_cluster_bias 	sv_size 
	------------------------------------------------------------------------
	CCS			5			10 bp (<500 bp)		20 bp 	
	CLR			5			20 bp (<500 bp)		50 bp 	
	------------------------------------------------------------------------
	
	Input file format
	------------------------------------------------------------------------
	column	#1	#2	#3	#4	#5
			INV	CHR	BP1	BP2	ID	
	#1	inversion type
	#2	chromosome number
	#3	breakpoint_1 in each read
	#4	breakpoint_2 in each read
	#5	read ID
	************************************************************************
	'''

	# Initialization of some temporary variables
	semi_inv_cluster = list()
	semi_inv_cluster.append([0,0,'',''])
	candidate_single_SV = list()

	# Load inputs & cluster breakpoint from each signature read 
	file = open(path, 'r')
	for line in file:
		seq = line.strip('\n').split('\t')
		if seq[1] != chr:
			continue

		strand = seq[2]
		breakpoint_1_in_read = int(seq[3])
		breakpoint_2_in_read = int(seq[4])
		read_id = seq[5]

		# print("new")
		# print(seq[1], seq[2], seq[3], seq[4], seq[5])
		# print(semi_inv_cluster)

		if breakpoint_1_in_read - semi_inv_cluster[-1][0] > max_cluster_bias or breakpoint_2_in_read - semi_inv_cluster[-1][1] > max_cluster_bias or strand != semi_inv_cluster[-1][-1]:
			if len(semi_inv_cluster) >= read_count:
				if semi_inv_cluster[-1][0] == semi_inv_cluster[-1][1] == 0:
					pass
				else:
					generate_semi_inv_cluster(semi_inv_cluster, 
											chr, 
											svtype, 
											read_count, 
											sv_size, 
											candidate_single_SV, 
											max_cluster_bias,
											bam_path,
											action,
											MaxSize,
											gt_round)
			semi_inv_cluster = []
			semi_inv_cluster.append([breakpoint_1_in_read, breakpoint_2_in_read, read_id, strand])
		else:
			semi_inv_cluster.append([breakpoint_1_in_read, breakpoint_2_in_read, read_id, strand])

	if len(semi_inv_cluster) >= read_count:
		if semi_inv_cluster[-1][0] == semi_inv_cluster[-1][1] == 0:
			pass
		else:
			generate_semi_inv_cluster(semi_inv_cluster, 
									chr, svtype, 
									read_count, 
									sv_size, 
									candidate_single_SV, 
									max_cluster_bias,
									bam_path,
									action,
									MaxSize,
									gt_round)
	file.close()
	return candidate_single_SV

def generate_semi_inv_cluster(semi_inv_cluster, chr, svtype, read_count, sv_size, 
	candidate_single_SV, max_cluster_bias, bam_path, action, MaxSize, gt_round):

	strand = semi_inv_cluster[0][-1]

	read_id = [i[2] for i in semi_inv_cluster]
	support_read = len(list(set(read_id)))
	if support_read < read_count:
		return

	inv_cluster_b2 = sorted(semi_inv_cluster, key = lambda x:x[1])

	# breakpoint_1 = np.mean(breakpoint_1_candidate)
	last_bp = inv_cluster_b2[0][1]
	temp_count = 1
	# max_count = 0
	temp_sum_b1 = inv_cluster_b2[0][0]
	temp_sum_b2 = last_bp

	# max_sum = 0
	temp_id = dict()
	temp_id[inv_cluster_b2[0][2]] = 0

	for i in inv_cluster_b2[1:]:
		if i[1] - last_bp > max_cluster_bias:
			if temp_count >= read_count:
				max_count_id = len(temp_id)

				breakpoint_1 = round(temp_sum_b1 / temp_count)
				breakpoint_2 = round(temp_sum_b2 / temp_count)
				inv_len = breakpoint_2 - breakpoint_1
				if inv_len >= sv_size and max_count_id >= read_count:
					# candidate_single_SV.append('%s\t%s\t%d\t%d\t%d\n'%(chr, svtype, breakpoint_1, breakpoint_2, max_count_id))
					if inv_len <= MaxSize:
						if action:
							import time
							# time_start = time.time()
							DV, DR, GT, GL, GQ, QUAL = call_gt(bam_path, int(breakpoint_1), 
															int(breakpoint_2), chr, list(temp_id.keys()), 
															max_cluster_bias, gt_round)
							# print(DV, DR, GT, GL, GQ, QUAL)
							# cost_time = time.time() - time_start
							# print("INV", chr, int(breakpoint_1), int(breakpoint_2), DR, DV, QUAL, "%.4f"%cost_time)
						else:
							DR = '.'
							GT = './.'
							GL = '.,.,.'
							GQ = "."
							QUAL = "."
						candidate_single_SV.append([chr, 
													svtype, 
													str(int(breakpoint_1)), 
													str(int(inv_len)), 
													str(max_count_id),
													str(DR),
													str(GT),
													strand,
													str(GL),
													str(GQ),
													str(QUAL),
													str(','.join(list(temp_id.keys())))])
						# print(chr, svtype, str(int(breakpoint_1)), str(int(inv_len)), str(max_count_id), str(DR), str(GT), strand)

			temp_id = dict()
			temp_count = 1
			temp_sum_b1 = i[0]
			temp_sum_b2 = i[1]
			temp_id[i[2]] = 0
		else:
			if i[2] not in temp_id:
				temp_id[i[2]] = 0
			else:
				temp_id[i[2]] += 1
			temp_count += 1
			temp_sum_b1 += i[0]
			temp_sum_b2 += i[1]
		last_bp = i[1]
	if temp_count >= read_count:
		max_count_id = len(temp_id)
		breakpoint_1 = round(temp_sum_b1 / temp_count)
		breakpoint_2 = round(temp_sum_b2 / temp_count)
		inv_len = breakpoint_2 - breakpoint_1
		if inv_len >= sv_size and max_count_id >= read_count:
			# candidate_single_SV.append('%s\t%s\t%d\t%d\t%d\n'%(chr, svtype, breakpoint_1, breakpoint_2, max_count_id))
			if inv_len <= MaxSize:
				if action:
					import time
					# time_start = time.time()
					DV, DR, GT, GL, GQ, QUAL = call_gt(bam_path, int(breakpoint_1), 
													int(breakpoint_2), chr, list(temp_id.keys()), 
													max_cluster_bias, gt_round)
					# print(DV, DR, GT, GL, GQ, QUAL)
					# cost_time = time.time() - time_start
					# print("INV", chr, int(breakpoint_1), int(breakpoint_2), DR, DV, QUAL, "%.4f"%cost_time)
				else:
					DR = '.'
					GT = './.'
					GL = '.,.,.'
					GQ = "."
					QUAL = "."
				candidate_single_SV.append([chr, 
											svtype, 
											str(int(breakpoint_1)), 
											str(int(inv_len)), 
											str(max_count_id),
											str(DR),
											str(GT),
											strand,
											str(GL),
											str(GQ),
											str(QUAL),
											str(','.join(list(temp_id.keys())))])
				# print(chr, svtype, str(int(breakpoint_1)), str(int(inv_len)), str(max_count_id), str(DR), str(GT), strand)

def run_inv(args):
	return resolution_INV(*args)

def call_gt(bam_path, pos_1, pos_2, chr, read_id_list, max_cluster_bias, gt_round):
	import pysam
	bamfile = pysam.AlignmentFile(bam_path)
	querydata = set()
	search_start = max(int(pos_1) - max_cluster_bias/2, 0)
	search_end = min(int(pos_1) + max_cluster_bias/2, bamfile.get_reference_length(chr))

	up_bound = threshold_ref_count(len(read_id_list))
	status = count_coverage(chr, 
							search_start, 
							search_end, 
							bamfile, 
							querydata, 
							up_bound, 
							gt_round)

	if status == -1:
		DR = '.'
		GT = "./."
		GL = ".,.,."
		GQ = "."
		QUAL = "."

	elif status == 1:
		DR = 0
		for query in querydata:
			if query not in read_id_list:
				DR += 1
		GT, GL, GQ, QUAL = cal_GL(DR, len(read_id_list))

	else:
		search_start = max(int(pos_2 - max_cluster_bias/2), 0)
		search_end = min(int(pos_2 + max_cluster_bias/2), bamfile.get_reference_length(chr))
		status_2 = count_coverage(chr, 
									search_start, 
									search_end, 
									bamfile, 
									querydata, 
									up_bound, 
									gt_round)
		# status_2 judgement
		DR = 0
		for query in querydata:
			if query not in read_id_list:
				DR += 1
		GT, GL, GQ, QUAL = cal_GL(DR, len(read_id_list))

	bamfile.close()
	return len(read_id_list), DR, GT, GL, GQ, QUAL
	
